#include "sobel.h"

int main(int argc, char *argv[])
{
	ImMain *ImStruct;
	int status;

	ImStruct = malloc(sizeof(ImMain));
	status = ParseInputs(argc, argv, ImStruct);
	status = InitDataStruct(ImStruct);

	status = ImageFilter(ImStruct);

	status = WriteImage(ImStruct);
	status = FreeDataStruct(ImStruct);

	return(0);
}

int InitDataStruct(ImMain *ImStruct)
{
	FILE *fd;

	/* read image size */
	if ((fd = fopen(ImStruct->ImFileName, "rb")) == NULL) {
		printf("Can't open Image file '%s'\n", ImStruct->ImFileName);
		exit(-1);
	}

	fseek(fd, 18, SEEK_SET);
	fread(&ImStruct->HorImSize, sizeof(int), 1, fd);
	fread(&ImStruct->VertImSize, sizeof(int), 1, fd);

	ImStruct->HeaderPtr = (unsigned char *)malloc(1078*sizeof(unsigned char));

	ImStruct->InPixelPtr = (unsigned char *)malloc(ImStruct->HorImSize*ImStruct->VertImSize*sizeof(unsigned char));
	ImStruct->OutPixelPtr = (unsigned char *)malloc(ImStruct->HorImSize*ImStruct->VertImSize*sizeof(unsigned char));

	ImStruct->ShortPixelPtrWide = (short *)malloc((ImStruct->HorImSize+2)*(ImStruct->VertImSize+2)*sizeof(short));
	ImStruct->Gxy = (short *)malloc((ImStruct->HorImSize)*(ImStruct->VertImSize) * sizeof(short));

	rewind(fd);

	fread(ImStruct->HeaderPtr, 1, 1078, fd);
	fread(ImStruct->InPixelPtr, 1, ImStruct->HorImSize*ImStruct->VertImSize, fd);

	fclose(fd);

	ImStruct->SobelX[0][0] = -1;
	ImStruct->SobelX[0][1] =  0;
	ImStruct->SobelX[0][2] =  1;
	ImStruct->SobelX[1][0] = -2;
	ImStruct->SobelX[1][1] =  0;
	ImStruct->SobelX[1][2] =  2;
	ImStruct->SobelX[2][0] = -1;
	ImStruct->SobelX[2][1] =  0;
	ImStruct->SobelX[2][2] =  1;

	ImStruct->SobelY[0][0] = -1;
	ImStruct->SobelY[0][1] = -2;
	ImStruct->SobelY[0][2] = -1;
	ImStruct->SobelY[1][0] =  0;
	ImStruct->SobelY[1][1] =  0;
	ImStruct->SobelY[1][2] =  0;
	ImStruct->SobelY[2][0] =  1;
	ImStruct->SobelY[2][1] =  2;
	ImStruct->SobelY[2][2] =  1;

	return(0);
}

void Usage(void)
{
	printf("Usage : sobel [-h] nom_du_fichier\n");
	exit(-1);
}

int ParseInputs(int argc, char *argv[], ImMain *ImStruct)
{
	int i, len;

	if (argc < 2) 
		Usage();
	else {
		for(i = 1; i < argc; i++) {
			if (strcmp(argv[i], "-h") == 0) 
				Usage();
			else
				strcpy(ImStruct->ImFileName, argv[i]);
		}
	}

	len = strlen(ImStruct->ImFileName);
	strncpy(ImStruct->OutImFileName, ImStruct->ImFileName, len - 4);
	ImStruct->OutImFileName[len - 4] = 0;
	strcat(ImStruct->OutImFileName, "_filt.bmp");

	return(0);
}

int WriteImage(ImMain *ImStruct)
{
	FILE *fd;

	/* read image size */
	if ((fd = fopen(ImStruct->OutImFileName, "wb")) == NULL) {
		printf("Can't open Image file '%s'\n", ImStruct->OutImFileName);
		exit(-1);
	}


	fwrite(ImStruct->HeaderPtr, 1, 1078, fd);
	fwrite(ImStruct->OutPixelPtr, 1, ImStruct->HorImSize*ImStruct->VertImSize, fd);

	fclose(fd);

	return(0);
}

int ImageFilter(ImMain *ImStruct)
{
	int i, j, u, v, H, V, mod_carre, sum_prodX, sum_prodY;
	short module, max;
	short *ptrs = ImStruct->ShortPixelPtrWide, *ptrsXY = ImStruct->Gxy;
	unsigned char *ptrci = ImStruct->InPixelPtr, *ptrco = ImStruct->OutPixelPtr;

	H = ImStruct->HorImSize;
	V = ImStruct->VertImSize;

	// copie image char vers short + ajout 2 colonnes, 2 lignes
	for (i = 0; i < V; i++) {
		for (j = 0; j < H; j++) {
			*(ptrs + (V-1-i)*(H+2) + j) = *(ptrci + i*H + j); // copie image avec inversion des lignes
		}
		*(ptrs + (V-1-i)*(H+2) + H) = *(ptrci + i*H + H-1); // extension 2 dernires colonnes
		*(ptrs + (V-1-i)*(H+2) + H + 1) = *(ptrci + i*H + H-1);
	}

	for (j = 0; j < H + 2; j++) { // extension 2 dernires lignes
		*(ptrs + V*(H + 2) + j) = *(ptrs + (V-1)*(H + 2) + j);
		*(ptrs + (V+1)*(H + 2) + j) = *(ptrs + (V-1)*(H + 2) + j);
	}

	// calcul image filtre horizontalement et verticalement

	max = 0;

	for (i = 0; i < V; i++) {
		for (j = 0; j < H; j++) {

			sum_prodX = 0;
			sum_prodY = 0;

			for (u = 0; u < 3; u++) {
				for (v = 0; v < 3; v++) {
					sum_prodX += (*(ptrs + (i + u)*(H + 2) + j + v)) * (ImStruct->SobelX[u][v]);
					sum_prodY += (*(ptrs + (i + u)*(H + 2) + j + v)) * (ImStruct->SobelY[u][v]);
				}
			}

			mod_carre = sum_prodX*sum_prodX + sum_prodY*sum_prodY;
			module = (short)sqrt((double)mod_carre);
			*(ptrsXY + i*H + j) = module;

			if (module > max) // calcul du pixel max
				max = module;
		}
	}

	// mise  l'chelle
	for (i = 0; i < V; i++) {
		for (j = 0; j < H; j++) {
			*(ptrsXY + i*H + j) = (short)(255.0 * (double)(*(ptrsXY + i*H + j)) / (double)max);
		}
	}



	// copie image filtre short vers char HxV
	for (i = 0; i < V; i++) {
		for (j = 0; j < H; j++) {
			*(ptrco + (V - 1 - i)*H + j) = (unsigned char)(*(ptrsXY + i*H + j)); // copie image avec inversion des lignes
		}
	}

	return(0);
}

int FreeDataStruct(ImMain *ImStruct)
{

	free(ImStruct->HeaderPtr);

	free(ImStruct->InPixelPtr);
	free(ImStruct->OutPixelPtr);

	free(ImStruct->ShortPixelPtrWide);
	free(ImStruct->Gxy);

	return(0);
}

